/*********************************************************/
/* UI_GLOBALS                                            */
/*   This base class is inherited by the button, menu,   */
/*   sub_window etc classes, and provides a way of       */
/*   safely accessing variables that need to be global   */
/*   to the entire user interface.                       */
/*                                                       */
/*                      (c)1994 Alexis 'Milamber' Ashley */
/*********************************************************/
/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __GLOBAL__
#define __GLOBAL__

#include "../globals/typedefs.H"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

enum corner {topleft, topright, bottomleft,bottomright };
typedef float world_c;

class UI_Globals
{
public:
// *** Public enums ***

  enum status {okay,out_of_memory,bad_parameter,unknown_error,EXIT,ABORT};

// *** Public structures ***

  struct opt_list           // This structure is used when checking for
    {                       // command line or resource options
      char *name;
      char *value;
    };

// *** Public functions ***

  UI_Globals(UI_Globals *parent);   //Tells UI_Globals who the parent window is
  UI_Globals(const char *const AppName,int argc,char **argv);
                // Tells UI_Globals the app's name & it's command line
  virtual ~UI_Globals(void);        // Destructor
  Display *DispPointer(void) {      // Returns the display pointer
    return(DisplayPointer);
  }
  UI_Globals *Parent(void) {        // Returns a pointer to the parent window
    return(This_Context.Parent);
  }
  status EventHandler(void);            // This is the main event handler.
                                        // It will run until one of the
                                        // event_handlers it calls returns EXIT
  status FlushEvents(void);             // This will service all events that
                                        // are pending, and then return.
  static const float MAX_WORLD_C_VAL;   // Maximum world coord value

  unsigned int x_spacing(void) {               // Returns current x spacing
    return(_x_spacing);
  }
  unsigned int y_spacing(void) {               // Returns current y spacing
    return(_y_spacing);
  }
  void RestrictEvents(Window valid_wins[], int win_count);
                                        // This will restrict all but expose
                                        // events to the list of windows
                                        // passed in valid_wins
  void UnrestrictedEvents(void);        // This undoes the previous function,
                                        // by allowing all events to be
                                        // processed.
  XErrorEvent GetLastError(void);       // Returns the last XErrorEvent struct
// **** All functions from here down cannot be called until the window ****
// **** has called RegisterWindow.                                     ****

  Window WinId(void) {                  // Returns the win id of this window
    return(_win_id);
  }
  unsigned int height(void) {           // Returns height of window
    return(_height);
  }
  unsigned int width(void) {            // Returns width of window
    return(_width);
  }
  world_c Height(void);                 // Returns height as a world coord
  world_c Width(void);                  // Returns width as a world coord
  unsigned int wc2pix_x(world_c);       // Converts a world coord to pixels
  unsigned int wc2pix_y(world_c);       // Converts a world coord to pixels
  unsigned int wc2pix_h(world_c);       // Converts a world coord to pixels
  unsigned int wc2pix_w(world_c);       // Converts a world coord to pixels
  world_c pix_x2wc(unsigned int);       // Converts a pixel to world coords
  world_c pix_y2wc(unsigned int);       // Converts a pixel to world coords
  world_c pix_h2wc(unsigned int);       // Converts a pixel to world coords
  world_c pix_w2wc(unsigned int);       // Converts a pixel to world coords
  void CheckResources(opt_list *);      // This function checks for the X
                                        // resources specified in res_list.
  void CheckCommandLine(opt_list *);    // This function checks for
                                        // options on the command line.
protected:
// *** Protected structures ***
  struct X_Context
    {
      UI_Globals *Parent;    // Points to the derrived class of the parent win
      UI_Globals *This;      // Points to the derrived class of this window
    };

// *** Protected static vars ***

  static XContext Contexts;
  static Colormap Colourmap;
  static colour Black;
  static colour White;

// *** Protected functions ***

  void RegisterWindow(UI_Globals *class_ptr, Window win, unsigned int h, 
		      unsigned int w);
                                        // This function is used to register
                                        // the size of a window, so that the
                                        // world coord conversions can be done
                                        // It also creates the X_Context struct
                                        // so that FindEvent can find this
                                        // window when it receives an event
  void ResizeWindow(unsigned int new_h, unsigned int new_w); // Resizes the window
  virtual status event_handler(XEvent *)=0;
private:
// *** Private Vars ***
  bool registered;             // =true if RegisterWindow has been called
  X_Context This_Context;      // The X_Context structure for this window
  float x_multiplier;          // Multiplier to convert world coord to pixels
  float y_multiplier;          // Multiplier to convert world coord to pixels
  Window _win_id;              // Registered window id
  unsigned int _width;         // Registered width of window
  unsigned int _height;        // Registered height of window
  Window *accept_wins;         // This is an array of all the windows
                               // that are to be allowed to handle all XEvents
  int a_w_count;               // This holds the size of the array
// *** Private static vars ***
  static int count;            // Count of the no. of instances of this class
  static unsigned int _x_spacing; //This stores the horiz spacing between items
  static unsigned int _y_spacing; // This stores the vertical spacing
  static Display *DisplayPointer; // This is the display pointer, which
                                  // is a bit like a file pointer,
                                  // but is used when talking to X
  static char *app_name;       // The name of the application
  static int argc;             // The number of command line options
  static char **argv;          // A pointer to the command line options
// *** Private Functions ***
  status FindEvent(XEvent *);  // Finds the event_handler & calls it
};

#endif

