/*********************************************************/
/* MENU                                                  */
/*   This class provides a popup menu system, which      */
/*   consists of a menu bar at the top of the parent     */
/*   window.  When an option is selected from this bar   */
/*   another window pops up, giving a list of the        */
/*   options for this menu.  The respective callback     */
/*   is called if one of the options is selected.  If    */
/*   the user clicks in any other window, the popup is   */
/*   closed.                                             */
/*                                                       */
/*                      (c)1994 Alexis 'Milamber' Ashley */
/*********************************************************/
/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __MENU__
#define __MENU__

#include "main_win.H"

class Menu : public UI_Globals
{
public:
  Menu(MainWindow *parent,const char * const names[20]); //The constructor
                    // You pass it the name of the window that you want the
                    // menu to be placed in, and the names of the menu
                    // options eg "File","Edit","Help" etc
  ~Menu(void);      // The destructor function

// Attach will attach a callback to a specified menu position, and return
// a MenuID structure, which you must keep, and use whenever you need
// to refer to the item again

  struct MenuID { int position, depth; };
  MenuID *attach(int position,const char * const label, char hotkey,
	 status (*callback)(byte *), byte *callback_data=NULL,
	 bool enabled=true);

  int Enable_Option(const MenuID * const id);  // This will enable an option
  int Disable_Option(const MenuID * const id); // This will disable an option
  void ResizeWindow(void);    // Resizes to size of main window
private:
// *** Private structures ***
  struct option            //This struct is used to store the options that
    {                      //popup when a toplevel menu item is selected
      char *name;
      char hotkey;
      bool enabled;
      status (*callback)(byte *);
      byte *callback_data;
      option *next;
    };
  struct menu_item
    {
      char *name;                  // Name of menu item Eg "File"
      unsigned int x1_pos;         // Left hand pixel of the menu item
      unsigned int x2_pos;         // Right hand pixel of the menu item
      option *first_option;        // Pointer to first option in the menu item
                                   // eg "Open"
    } m_item[20];                  // Array of 20 menu items 
                                   // (eg File, Edit, Help)
// *** Private vars ***

  GC GraphicsContext;            // The GC used for drawing in the menu
  GC OffGraphicsContext;         // The GC used for drawing disabled options
  colour fg_colour;              // The foreground colour
  colour bg_colour;              // The background colour
  colour off_colour;             // The colour used for drawing disabled opts
  XFontStruct *Menu_Font;        // Font used for drawing menu names & options
  int active_option;             // Current active option (-1=none active)
  int submenu_exit_var;          // This is set non-zero to stop a popup window
  X_Context PopupContext;        // This X_Context is used when there is an
                                 // active popup menu window open
  Window popup_win_id;           // The window id of the popup window
// *** Private static vars ***

  static char Foreground[80];     // The name of the foreground colour
  static char Background[80];     // The name of the background colour
  static char DisabledColour[80]; // The name of the colour for disabled opts
  static char FontName[80];       // The name of the menu's font
  static const unsigned int BorderWidth; // The borderwidth of the menu

// *** Private Functions ***

  status SubMenu_handler(int pos);
  status event_handler(XEvent *);// The event handler, called by UI_Globals
                     // It passes control to one of the two below handlers
  status MainMenu_event_handler(XEvent *); //This is used when there's no popup
  status SubMenu_event_handler(XEvent *);  //This is used when there is a popup
};

#endif 
